#!/usr/bin/env python3
"""
SC-Controller - Action Editor - Button Component

Assigns emulated button to physical button
"""
from scc.tools import _

from gi.repository import Gtk, Gdk, GLib
from scc.actions import Action, ButtonAction, MouseAction
from scc.actions import AxisAction, MultiAction, NoAction
from scc.macros import Macro, Cycle, PressAction, ReleaseAction
from scc.uinput import Rels, Keys
from scc.gui.area_to_action import action_to_area
from scc.gui.key_grabber import KeyGrabber
from scc.gui.parser import InvalidAction
from scc.gui.chooser import Chooser
from scc.gui.ae import AEComponent

import os, logging
log = logging.getLogger("AE.Buttons")

__all__ = [ 'ButtonsComponent' ]


class ButtonsComponent(AEComponent, Chooser):
	GLADE = "ae/buttons.glade"
	NAME = "buttons"
	IMAGES = { "buttons" : "buttons.svg" }
	CTXS = Action.AC_BUTTON | Action.AC_MENU
	PRIORITY = 1
	MODIFIER_KEYS = ( Keys.KEY_LEFTSHIFT, Keys.KEY_LEFTMETA, Keys.KEY_LEFTALT,
		Keys.KEY_LEFTCTRL, Keys.KEY_RIGHTMETA, Keys.KEY_RIGHTSHIFT,
		Keys.KEY_RIGHTCTRL, Keys.KEY_RIGHTALT,
	)

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		Chooser.__init__(self, app)
		self.axes_allowed = True
		self.keys = set()


	def load(self):
		if not self.loaded:
			AEComponent.load(self)
			self.setup_image()
			if self.app.osd_mode:
				self.builder.get_object('btnGrabKey').set_sensitive(False)
				self.builder.get_object('btnGrabAnother').set_sensitive(False)


	def area_action_selected(self, area, action):
		self.set_active_area(area)
		self.editor.set_action(action)


	def set_action(self, mode, action):
		cbToggle = self.builder.get_object("cbToggle")
		cbRepeat = self.builder.get_object("cbRepeat")
		if self.handles(mode, action):
			self.keys = set()
			is_togle, is_repeat = False, False
			if isinstance(action, MultiAction):
				for a in action.actions:
					if isinstance(a, ButtonAction):
						self.keys.add(a.button)
			elif isinstance(action, ButtonAction):
				self.keys.add(action.button)
			elif isinstance(action, Macro):
				# Macro goes here only if it is button repeat
				self.keys.add(action.actions[0].button)
				is_repeat = True
			elif isinstance(action, Cycle):
				# There is only one case when self.handles returns True for Cycle
				self.keys.add(action.actions[0].action.button)
				is_togle = True
			cbToggle.set_active(is_togle)
			cbRepeat.set_active(is_repeat)
			area = action_to_area(action)
			if area is not None:
				self.set_active_area(area)
				return
		self.set_active_area(None)


	def get_button_title(self):
		return _("Key or Button")


	def handles(self, mode, action):
		# Handles ButtonAction and MultiAction if all subactions are ButtonAction
		if isinstance(action, (ButtonAction, NoAction, InvalidAction)):
			return True
		if isinstance(action, AxisAction):
			return len(action.parameters) == 1
		if isinstance(action, MouseAction):
			if action.get_axis() == Rels.REL_WHEEL:
				return True
		if is_button_togle(action):
			return True
		if is_button_repeat(action):
			return True
		if isinstance(action, MultiAction):
			if len(action.actions) > 0:
				for a in action.actions:
					if not isinstance(a, ButtonAction):
						return False
				return True
		return False


	def on_key_grabbed(self, keys):
		""" Handles selecting key using "Grab the Key" dialog """
		self.keys = set(keys)
		self.apply_keys()


	def on_additional_key_grabbed(self, keys):
		self.keys.update(keys)
		self.apply_keys()


	@staticmethod
	def modifiers_first(key):
		if key in ButtonsComponent.MODIFIER_KEYS:
			return 0
		return 1


	def apply_keys(self, *a):
		""" Common part of on_*key_grabbed """
		cbToggle = self.builder.get_object("cbToggle")
		cbRepeat = self.builder.get_object("cbRepeat")
		keys = list(sorted(self.keys, key=ButtonsComponent.modifiers_first))
		action = ButtonAction(keys[0])
		if len(keys) > 1:
			actions = [ ButtonAction(k) for k in keys ]
			action = MultiAction(*actions)
		if cbRepeat.get_active():
			action = Macro(action)
			action.repeat = True
		elif cbToggle.get_active():
			action = Cycle(PressAction(action), ReleaseAction(action))
		self.editor.set_action(action)


	def on_btnGrabKey_clicked(self, *a):
		"""
		Called when user clicks on 'Grab a Key' button.
		Displays additional dialog.
		"""
		kg = KeyGrabber(self.app)
		kg.grab(self.editor.window, self.editor._action, self.on_key_grabbed)


	def on_btnGrabAnother_clicked(self, *a):
		"""
		Same as above, but adds another key to action
		"""
		kg = KeyGrabber(self.app)
		kg.grab(self.editor.window, self.editor._action,
				self.on_additional_key_grabbed)


	def on_cbToggle_toggled(self, cbToggle):
		cbRepeat = self.builder.get_object("cbRepeat")
		if cbToggle.get_active() and cbRepeat.get_active():
			cbRepeat.set_active(False)
		self.apply_keys()


	def on_cbRepeat_toggled(self, cbRepeat):
		cbToggle = self.builder.get_object("cbToggle")
		if cbToggle.get_active() and cbRepeat.get_active():
			cbToggle.set_active(False)
		self.apply_keys()


	def hide_toggle(self):
		""" Hides 'set as toggle button' option """
		cbToggle = self.builder.get_object("cbToggle")
		btnGrabAnother = self.builder.get_object("btnGrabAnother")
		replacement = Gtk.Label("")
		replacement.set_size_request(*cbToggle.get_size_request())
		btnGrabAnother.set_visible(False)
		parent = cbToggle.get_parent()
		parent.remove(cbToggle)
		parent.pack_start(replacement, True, True, 0)
		parent.reorder_child(replacement, 0)
		replacement.set_visible(True)


	def hide_axes(self):
		""" Prevents user from selecting axes """
		self.axes_allowed = False


def is_button_togle(action):
	if not isinstance(action, Cycle):
		return False
	if len(action.actions) != 2:
		return False
	if isinstance(action.actions[0], PressAction):
		if isinstance(action.actions[1], ReleaseAction):
			if isinstance(action.actions[0].action, ButtonAction):
				if isinstance(action.actions[1].action, ButtonAction):
					return action.actions[0].action.button == action.actions[1].action.button
	return False


def is_button_repeat(action):
	if isinstance(action, Cycle):
		return False
	if isinstance(action, Macro) and action.repeat:
		if len(action.actions) == 1:
			return isinstance(action.actions[0], ButtonAction)
	return False



import ctypes
import fcntl
from typing import NamedTuple

import ioctl_opt

# input.h
BUS_USB = 0x03
BUS_HIL = 0x04
BUS_BLUETOOTH = 0x05
BUS_VIRTUAL = 0x06

# hid.h
_HID_MAX_DESCRIPTOR_SIZE = 4096

# hidraw.h
class _hidraw_report_descriptor(ctypes.Structure):
    _fields_ = [
        ('size', ctypes.c_uint),
        ('value', ctypes.c_ubyte * _HID_MAX_DESCRIPTOR_SIZE),
    ]

class _hidraw_devinfo(ctypes.Structure):
    _fields_ = [
        ('bustype', ctypes.c_uint),
        ('vendor', ctypes.c_short),
        ('product', ctypes.c_short),
    ]

_HIDIOCGRDESCSIZE = ioctl_opt.IOR(ord('H'), 0x01, ctypes.c_int)
_HIDIOCGRDESC = ioctl_opt.IOR(ord('H'), 0x02, _hidraw_report_descriptor)
_HIDIOCGRAWINFO = ioctl_opt.IOR(ord('H'), 0x03, _hidraw_devinfo)
_HIDIOCGRAWNAME = lambda len: ioctl_opt.IOC(ioctl_opt.IOC_READ, ord('H'),
    0x04, len)
_HIDIOCGRAWPHYS = lambda len: ioctl_opt.IOC(ioctl_opt.IOC_READ, ord('H'),
    0x05, len)
_HIDIOCSFEATURE = lambda len: ioctl_opt.IOC(
    ioctl_opt.IOC_WRITE|ioctl_opt.IOC_READ, ord('H'), 0x06, len)
_HIDIOCGFEATURE = lambda len: ioctl_opt.IOC(
    ioctl_opt.IOC_WRITE|ioctl_opt.IOC_READ, ord('H'), 0x07, len)

HIDRAW_FIRST_MINOR = 0
HIDRAW_MAX_DEVICES = 64
HIDRAW_BUFFER_SIZE = 64

class DevInfo(NamedTuple):
	"""Device Info.

	- bustype: one of BUS_USB, BUS_HIL, BUS_BLUETOOTH or BUS_VIRTUAL
	- vendor: device's vendor number
	- product: device's product number
"""

	bustype: int
	vendor: int
	product: int

class HIDRaw:
    """Provide methods to access hidraw device's ioctls."""

    def __init__(self, device):
        """device (file, fileno).

            A file object or a fileno of an open hidraw device node.
        """
        self._device = device

    def _ioctl(self, func, arg, mutate_flag: bool = False):
        result = fcntl.ioctl(self._device, func, arg, mutate_flag)
        if result < 0:
            raise OSError(result)

    def read(self, size):
        return self._device.read(size)

    def write(self, buf):
        return self._device.write(buf)

    def getRawReportDescriptor(self) -> str:
        """Return a binary string containing the raw HID report descriptor."""
        descriptor = _hidraw_report_descriptor()
        size = ctypes.c_uint()
        self._ioctl(_HIDIOCGRDESCSIZE, size, True)
        descriptor.size = size
        self._ioctl(_HIDIOCGRDESC, descriptor, True)
        return "".join(chr(x) for x in descriptor.value[:size.value])

    # TODO: decode descriptor into a python object
    #def getReportDescriptor(self):

    def getInfo(self) -> DevInfo:
        """Return a DevInfo instance."""
        devinfo = _hidraw_devinfo()
        self._ioctl(_HIDIOCGRAWINFO, devinfo, True)
        return DevInfo(devinfo.bustype, devinfo.vendor, devinfo.product)

    def getName(self, length:int = 512) -> str:
        """Return device name as an unicode object."""
        name = ctypes.create_string_buffer(length)
        self._ioctl(_HIDIOCGRAWNAME(length), name, True)
        return name.value.decode("UTF-8")

    def getPhysicalAddress(self, length:int = 512) -> str:
        """Return device's physical address as a string.

        See hidraw documentation for value signification, as it depends on
        device's bus type.
        """
        name = ctypes.create_string_buffer(length)
        self._ioctl(_HIDIOCGRAWPHYS(length), name, True)
        return name.value

    def sendFeatureReport(self, report, report_num:int = 0) -> None:
        """Send a feature report."""
        length = len(report) + 1
        buf = bytearray(length)
        buf[0] = report_num
        buf[1:] = report
        self._ioctl(
            _HIDIOCSFEATURE(length),
            (ctypes.c_char * length).from_buffer(buf),
            True,
        )

    def getFeatureReport(self, report_num:int = 0, length:int = 63) -> bytearray:
        """Receive a feature report.

        Blocks, unless you configured provided file (descriptor) to be
        non-blocking.
        """
        length += 1
        buf = bytearray(length)
        buf[0] = report_num
        self._ioctl(
            _HIDIOCGFEATURE(length),
            (ctypes.c_char * length).from_buffer(buf),
            True,
        )
        return buf



#!/usr/bin/env python3
"""
SC-Controller - Gestures

Everything related to non-GUI part of gesture detection lies here.
It's technically part of SCC-Daemon, separater into special module just to keep
it clean.
"""
from scc.actions import Action
from scc.tools import circle_to_square, clamp
from scc.constants import STICK_PAD_MIN, STICK_PAD_MAX, CPAD, CPAD_MIN
from scc.constants import CPAD_X_MAX, CPAD_Y_MAX
from math import pi as PI, atan2, sqrt
from itertools import groupby

import logging
log = logging.getLogger("Gestures")


class GestureDetector(Action):
	"""
	Derived from Action, but not callable in profile.
	
	When daemon decides it's good time to start gesture, be it because of
	GestureAction special action or "Gesture:" message from client,
	it constructs instance of this class and leaves everything to it.
	"""
	UP			= "U"
	DOWN		= "D"
	LEFT		= "L" 
	RIGHT		= "R"
	
	
	def __init__(self, up_direction, on_finished):
		Action.__init__(self)
		# TODO: Configurable resolution
		self._resolution = 3
		self._deadzone = 1.0 / self._resolution / self._resolution
		self._up_direction = up_direction
		self._on_finished = on_finished
		self._enabled = False
		self._positions = []
		self._result = []
	
	
	def enable(self):
		""" GestureDetector doesn't starts do detect anything until this is called """
		self._enabled = True
		self._result = [ ]
	
	
	def get_string(self):
		""" Returns string representation of (probably unfinished) gesture """
		return "".join(self._result)
	
	
	def get_positions(self):
		""" Returns list of positions used to generate gesture """
		return self._positions
	
	
	def get_resolution(self):
		""" Returns gesture resolution """
		return self._resolution
	
	
	def whole(self, mapper, x, y, what):
		if self._enabled:
			if (x, y) == (0, 0):
				# Pad was released
				self._enabled = False
				self._on_finished(self, "".join(self._result))
				return
			else:
				# Convert positions on pad to position on grid
				if what == CPAD:
					x = clamp(0, float(x) / (CPAD_X_MAX - CPAD_MIN), 1.0)
					y = clamp(0, float(y) / (CPAD_Y_MAX - CPAD_MIN), 1.0)
					x *= self._resolution
					y *= self._resolution
				else:
					x -= STICK_PAD_MIN
					y = STICK_PAD_MAX - y
					x = float(x) / (float(STICK_PAD_MAX - STICK_PAD_MIN) / self._resolution)
					y = float(y) / (float(STICK_PAD_MAX - STICK_PAD_MIN) / self._resolution)
				# Check for deadzones around grid lines
				for i in range(1, self._resolution):
					if x > i - self._deadzone and x < i + self._deadzone: return
					if y > i - self._deadzone and y < i + self._deadzone: return
				# Round
				x = clamp(0, int(x), self._resolution - 1)
				y = clamp(0, int(y), self._resolution - 1)
				if self._positions:
					ox, oy = self._positions[-1]
					if (x, y) != (ox, oy):
						self._positions.append( (x, y) )
						while (x, y) != (ox, oy):
							if x < ox:
								self._result.append(self.LEFT)
								x += 1
							elif x > ox:
								self._result.append(self.RIGHT)
								x -= 1
							elif y < oy:
								self._result.append(self.UP)
								y += 1
							elif y > oy:
								self._result.append(self.DOWN)
								y -= 1
				else:
					self._positions.append( (x, y) )




#!/usr/bin/env python3
"""
SC-Controller - Action Editor

Allows to edit button or trigger action.
"""
from scc.tools import _

from scc.gui.controller_widget import ControllerButton
from scc.gui.controller_widget import STICKS, PADS
from scc.gui.dwsnc import headerbar
from scc.gui.editor import Editor
from scc.actions import Action, NoAction, RangeOP, RingAction
from scc.constants import SCButtons, HapticPos, TRIGGER_MAX
from scc.modifiers import ModeModifier, DoubleclickModifier
from scc.modifiers import FeedbackModifier, HoldModifier
from scc.profile import Profile
from scc.macros import Macro
from scc.tools import nameof

from gi.repository import Gtk, Gdk, GLib
import os, logging
log = logging.getLogger("ModeshiftEditor")


class ModeshiftEditor(Editor):
	GLADE = "modeshift_editor.glade"
	BUTTONS = (	# in order as displayed in combobox
		(SCButtons.A,			_('A') ),
		(SCButtons.B,			_('B') ),
		(SCButtons.X,			_('X') ),
		(SCButtons.Y,			_('Y') ),
		(None, None),
		(SCButtons.BACK,		_('Back (select)') ),
		(SCButtons.C,			_('Center') ),
		(SCButtons.START,		_('Start') ),
		(None, None),
		(SCButtons.LGRIP,		_('Left Grip') ),
		(SCButtons.RGRIP,		_('Right Grip') ),
		(SCButtons.LB,			_('Left Bumper') ),
		(SCButtons.RB,			_('Right Bumper') ),
		(None, None),
		(SCButtons.LT,			_('Left Trigger (full)') ),
		(SCButtons.RT,			_('Right Trigger (full)') ),
		("Soft LT",				_('Left Trigger (soft)') ),
		("Soft RT",				_('Right Trigger (soft)') ),
		(None, None),
		(SCButtons.STICKPRESS,	_('Stick Pressed') ),
		(SCButtons.LPAD,		_('Left Pad Pressed') ),
		(SCButtons.RPAD,		_('Right Pad Pressed') ),
		(SCButtons.LPADTOUCH,	_('Left Pad Touched') ),
		(SCButtons.RPADTOUCH,	_('Right Pad Touched') ),
	)

	def __init__(self, app, callback):
		Editor.__init__(self)
		self.app = app
		self.id = None
		self.mode = Action.AC_BUTTON
		self.ac_callback = callback
		self.soft_level = -1
		self.current_page = 0
		self.actions = ( [], [], [] )
		self.nomods = [ NoAction(), NoAction(), NoAction() ]
		self.setup_widgets()


	def setup_widgets(self):
		Editor.setup_widgets(self)

		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		cbButtonChooser.set_row_separator_func( lambda model, iter : model.get_value(iter, 0) is None )

		b = lambda a : self.builder.get_object(a)
		self.action_widgets = (
			# Order goes: Grid, 1st Action Button, Clear Button
			# 1st group, 'pressed'
			( b('grActions'),		b('btDefault'),		b('btClearDefault') ),
			# 2nd group, 'hold'
			( b('grHold'),			b('btHold'),		b('btClearHold') ),
			# 2nd group, 'double-click'
			( b('grDoubleClick'),	b('btDoubleClick'),	b('btClearDoubleClick') ),
		)

		headerbar(self.builder.get_object("header"))


	def on_Dialog_destroy(self, *a):
		self.remove_added_widget()


	def _fill_button_chooser(self, *a):
		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		model = cbButtonChooser.get_model()
		model.clear()
		for item, text in self.BUTTONS:
			if any([ x[0] == item for x in self.actions[self.current_page] ]):
				# Skip already added buttons
				continue
			if type(item) is str:
				# Special case for soft pull items
				button = getattr(SCButtons, item.split(" ")[-1])
				if any([ (isinstance(x, RangeOP) and x.what == button)
							for x in self.actions[self.current_page] ]):
					# Skip already added soft pulls
					continue

			if item == SCButtons.STICKPRESS:
				if self.id == nameof(SCButtons.LPAD):
					# Controller cannot handle pressing stick and lpad at once
					continue
			model.append(( None if item is None else nameof(item), text ))
		cbButtonChooser.set_active(0)


	def _add_action(self, index, what, action):
		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		adjSoftLevel = self.builder.get_object("adjSoftLevel")
		rvSoftLevel = self.builder.get_object("rvSoftLevel")
		grActions = self.action_widgets[index][0]
		model = cbButtonChooser.get_model()

		for row in model:
			item = model.get_value(row.iter, 0)
			if item:
				if item == nameof(what):
					model.remove(row.iter)
					break
				if isinstance(what, RangeOP) and item.startswith("Soft"):
					if nameof(what.what) == item.split(" ")[-1]:
						model.remove(row.iter)
						break
		try:
			while model.get_value(model[0].iter, 0) is None:
				model.remove(model[0].iter)
			cbButtonChooser.set_active(0)
		except: pass

		i = len(self.actions[index]) + 1
		l = Gtk.Label()
		if isinstance(what, RangeOP) and what.op == ">=":
			if self.soft_level == -1:
				# First Range added, load level from it
				self.soft_level = what.value if what.value > 0.0 else 0.75
				adjSoftLevel.set_value(self.soft_level)
				rvSoftLevel.set_reveal_child(True)
				what.value = -1
			if what.value == -1:
				# Range with value taken from "Soft Pull Level" slider
				l.set_markup("<b>%s (soft pull)</b>" % (nameof(what.what),))
			else:
				# Any other range
				l.set_markup("<b>%s</b>" % (nameof(what),))
		else:
			l.set_markup("<b>%s</b>" % (nameof(what),))
		l.set_xalign(0.0)
		b = Gtk.Button.new_with_label(action.describe(self.mode))
		b.set_property("hexpand", True)
		b.connect('clicked', self.on_actionb_clicked, index, what)
		clearb = Gtk.Button()
		clearb.set_image(Gtk.Image.new_from_stock("gtk-delete", Gtk.IconSize.SMALL_TOOLBAR))
		clearb.set_relief(Gtk.ReliefStyle.NONE)
		clearb.connect('clicked', self.on_clearb_clicked, index, what)
		grActions.attach(l,			0, i, 1, 1)
		grActions.attach(b,			1, i, 1, 1)
		grActions.attach(clearb,	2, i, 1, 1)

		self.actions[index].append([ what, action, l, b, clearb ])
		grActions.show_all()


	def on_clearb_clicked(self, trash, index, button):
		grActions = self.action_widgets[index][0]
		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		model = cbButtonChooser.get_model()
		# Remove requested action from the list
		for i in range(0, len(self.actions[index])):
			if self.actions[index][i][0] == button:
				button, action, l, b, clearb = self.actions[index][i]
				for w in (l, b, clearb): grActions.remove(w)
				del self.actions[index][i]
				break
		# Move everything after that action one position up
		# - remove it
		for j in range(i, len(self.actions[index])):
			button, action, l, b, clearb = self.actions[index][j]
			for w in (l, b, clearb): grActions.remove(w)
		# - add it again
		for j in range(i, len(self.actions[index])):
			button, action, l, b, clearb = self.actions[index][j]
			grActions.attach(l,			0, j + 1, 1, 1)
			grActions.attach(b,			1, j + 1, 1, 1)
			grActions.attach(clearb,	2, j + 1, 1, 1)
		# Regenereate combobox with removed button added back to it
		# - Store acive item from in combobox
		active, i, index = None, 0, -1
		try:
			active = model.get_value(cbButtonChooser.get_active_iter(), 0)
		except: pass
		# Clear entire combobox
		model.clear()
		# Fill it again
		for button, text in self.BUTTONS:
			model.append(( None if button is None else nameof(button), text ))
			if button is not None:
				if nameof(button) == active:
					index = i
			i += 1
		# Reselect formely active item
		if index >= 0:
			cbButtonChooser.set_active(index)


	def _choose_editor(self, action, cb):
		from scc.gui.ring_editor import RingEditor	# Cannot be imported @ top
		if isinstance(action, Macro):
			from scc.gui.macro_editor import MacroEditor	# Cannot be imported @ top
			e = MacroEditor(self.app, cb)
			e.set_title(_("Edit Macro"))
		elif RingEditor.is_ring_action(action):
			e = RingEditor(self.app, cb)
		else:
			from scc.gui.action_editor import ActionEditor	# Cannot be imported @ top
			e = ActionEditor(self.app, cb)
			e.set_title(_("Edit Action"))
			e.hide_modeshift()
		return e


	def on_actionb_clicked(self, trash, index, clicked_button):
		for i in self.actions[index]:
			button, action, l, b, clearb = i
			if button == clicked_button:
				def on_chosen(id, action):
					b.set_label(action.describe(self.mode))
					i[1] = action

				ae = self._choose_editor(action, on_chosen)
				ae.set_input(self.id, action, mode = self.mode)
				ae.show(self.window)
				return


	def on_ntbMore_switch_page(self, ntb, box, index):
		self.current_page = index
		self._fill_button_chooser()
		self.builder.get_object("cbButtonChooser").set_sensitive(box.get_sensitive())
		self.builder.get_object("btAddAction").set_sensitive(box.get_sensitive())


	def on_nomodbt_clicked(self, button, *a):
		actionButton = self.action_widgets[self.current_page][1]

		def on_chosen(id, action):
			actionButton.set_label(action.describe(self.mode))
			self.nomods[self.current_page] = action

		ae = self._choose_editor(self.nomods[self.current_page], on_chosen)
		ae.set_input(self.id, self.nomods[self.current_page], mode = self.mode)
		ae.show(self.window)


	def on_nomodclear_clicked(self, button, *a):
		self.nomods[self.current_page] = NoAction()
		actionButton = self.action_widgets[self.current_page][1]
		actionButton.set_label(self.nomods[self.current_page].describe(self.mode))


	def on_btAddAction_clicked(self, *a):
		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		item = cbButtonChooser.get_model().get_value(cbButtonChooser.get_active_iter(), 0)
		if item.startswith("Soft"):
			b = getattr(SCButtons, item.split(" ")[-1])
			rng = RangeOP(b, ">=", -1)
			self._add_action(self.current_page, rng, NoAction())
		else:
			b = getattr(SCButtons, item)
			self._add_action(self.current_page, b, NoAction())


	def on_sclSoftLevel_format_value(self, scale, value):
		return  "%s%%" % (int(value * 100.0),)


	def on_btClear_clicked(self, *a):
		""" Handler for clear button """
		action = NoAction()
		if self.ac_callback is not None:
			self.ac_callback(self.id, action)
		self.close()


	def on_btCustomActionEditor_clicked(self, *a):
		""" Handler for 'Custom Editor' button """
		from scc.gui.action_editor import ActionEditor	# Can't be imported on top
		e = ActionEditor(self.app, self.ac_callback)
		e.set_input(self.id, self._make_action(), mode = self.mode)
		e.hide_action_buttons()
		e.hide_advanced_settings()
		e.set_title(self.window.get_title())
		e.force_page(e.load_component("custom"), True)
		self.send_added_widget(e)
		self.close()
		e.show(self.get_transient_for())


	def on_cbHoldFeedback_toggled(self, cb, *a):
		rvHoldFeedbackAmplitude = self.builder.get_object("rvHoldFeedbackAmplitude")
		rvHoldFeedbackAmplitude.set_reveal_child(cb.get_active())


	def on_btOK_clicked(self, *a):
		""" Handler for OK button """
		if self.ac_callback is not None:
			self.ac_callback(self.id, self._make_action())
		self.close()


	def _make_action(self):
		""" Generates and returns Action instance """
		cbHoldFeedback = self.builder.get_object("cbHoldFeedback")
		sclHoldFeedback = self.builder.get_object("sclHoldFeedback")
		normalaction = self._save_modemod(0)
		holdaction = self._save_modemod(1)
		dblaction = self._save_modemod(2)
		if dblaction:
			action = DoubleclickModifier(dblaction, normalaction)
			action.holdaction = holdaction
		elif holdaction:
			action = HoldModifier(holdaction, normalaction)
		else:
			action = normalaction
		action.timeout = self.builder.get_object("adjTime").get_value()

		if cbHoldFeedback.get_active():
			action = FeedbackModifier(HapticPos.BOTH,
						sclHoldFeedback.get_value(), action)

		return action


	def _save_modemod(self, index):
		""" Generates ModeModifier from page in Notebook """
		adjSoftLevel = self.builder.get_object("adjSoftLevel")
		pars = []
		for item, action, l, b, clearb in self.actions[index]:
			if isinstance(item, RangeOP) and item.value == -1:
				# Value taken from slider
				item.value = adjSoftLevel.get_value()
			pars += [ item, action ]
		if self.nomods[index]:
			pars += [ self.nomods[index] ]
		action = ModeModifier(*pars)
		if len(pars) == 0:
			# No action is actually set
			action = NoAction()
		elif len(pars) == 1:
			# Only default action left
			action = self.nomods[index]
		return action


	def _load_modemod(self, index, action):
		for key in action.mods:
			self._add_action(index, key, action.mods[key])


	def _set_nomod_button(self, index, action):
		if isinstance(action, ModeModifier):
			self._load_modemod(index, action)
			self.nomods[index] = action.default
		else:
			self.nomods[index] = action
		actionButton = self.action_widgets[index][1]
		actionButton.set_label(self.nomods[index].describe(self.mode))


	def allow_first_page(self):
		""" For compatibility with action editor. Does nothing """
		pass


	def set_input(self, id, action, mode=None):
		btDefault = self.builder.get_object("btDefault")
		lblPressAlone = self.builder.get_object("lblPressAlone")
		cbHoldFeedback = self.builder.get_object("cbHoldFeedback")
		sclHoldFeedback = self.builder.get_object("sclHoldFeedback")

		self.id = id
		self._fill_button_chooser()

		if id in STICKS:
			lblPressAlone.set_label(_("(no button pressed)"))
			self.mode = mode = mode or Action.AC_STICK
		elif id in PADS:
			lblPressAlone.set_label(_("(no button pressed)"))
			self.mode = mode = mode or Action.AC_PAD
		else:
			lblPressAlone.set_label(_("(pressed alone)"))
			self.mode = mode = mode or Action.AC_BUTTON

		self.set_title("Modeshift for %s" % (nameof(id) if id in SCButtons.__members__.values() else str(id),))

		if isinstance(action, FeedbackModifier):
			cbHoldFeedback.set_active(True)
			sclHoldFeedback.set_value(action.haptic.get_amplitude())
			action = action.action
		else:
			cbHoldFeedback.set_active(False)
			sclHoldFeedback.set_value(512)

		if isinstance(action, ModeModifier):
			self._load_modemod(0, action)
			self._set_nomod_button(0, action.default)
			self._set_nomod_button(1, NoAction())
			self._set_nomod_button(2, NoAction())
		elif isinstance(action, DoubleclickModifier):	# includes HoldModifier
			self._set_nomod_button(0, action.normalaction)
			self._set_nomod_button(1, action.holdaction)
			self._set_nomod_button(2, action.action)
		self.builder.get_object("adjTime").set_value(action.timeout)

		if mode == Action.AC_OSK:
			# This is kinda bad, but allowing Custom Editor
			# from OSK editor is in TODO
			self.builder.get_object("btCustomActionEditor").set_visible(False)
		if mode != Action.AC_BUTTON:
			for w in ("vbHold", "vbDoubleClick", "lblHold", "lblDoubleClick"):
				self.builder.get_object(w).set_sensitive(False)



#!/usr/bin/env python3
"""
SC-Controller - Action Editor - OSK Button Component

Binds controller buttons on on on on on... screen keyboard.
Retuses ButtonsComponent, but hides image, so user can't select mouse or gamepad
button.
"""
from scc.tools import _

from gi.repository import Gtk, Gdk, GLib
from scc.actions import Action, ButtonAction, MultiAction, NoAction
from scc.gui.area_to_action import action_to_area
from scc.gui.ae.buttons import ButtonsComponent
from scc.gui.key_grabber import KeyGrabber
from scc.gui.parser import InvalidAction
from scc.gui.chooser import Chooser
from scc.gui.ae import AEComponent

import os, logging
log = logging.getLogger("AE.Buttons")

__all__ = [ 'OSKButtonsComponent' ]


class OSKButtonsComponent(ButtonsComponent):
	CTXS = Action.AC_OSK
	PRIORITY = 1
	IMAGES = { }


	def get_button_title(self):
		return _("Key")


	def load(self):
		if not self.loaded:
			AEComponent.load(self)
			self.builder.get_object("lblClickAnyButton").set_visible(False)



#!/usr/bin/env python3
"""
SC-Controller - OSD Dialog

Display dialog with text and set of items that user can navigate through and
prints chosen item id to stdout
"""
from gi.repository import Gtk, GdkX11
from scc.gui.daemon_manager import DaemonManager
from scc.osd import OSDWindow, StickController
from scc.lib import xwrappers as X
from scc.constants import DEFAULT, STICK
from scc.menu_data import MenuData
from scc.config import Config

import sys, logging
log = logging.getLogger("osd.dialog")


class Dialog(OSDWindow):
	EPILOG="""Exit codes:
   0  - clean exit, user selected option
  -1  - clean exit, user canceled dialog
   1  - error, invalid arguments
   2  - error, failed to access sc-daemon, sc-daemon reported error or died while dialog is displayed.
   3  - erorr, failed to lock input stick, pad or button(s)
	"""

	def __init__(self, cls="osd-menu"):
		self._buttons = None
		self._text = None

		OSDWindow.__init__(self, cls)
		self.daemon = None
		self.config = None
		self.feedback = None
		self.controller = None
		self.xdisplay = X.Display(hash(GdkX11.x11_get_default_xdisplay()))	# Magic

		self.parent = self.create_parent()
		self.f = Gtk.Fixed()
		self.f.add(self.parent)
		self.add(self.f)

		self._scon = StickController()
		self._scon.connect("direction", self.on_stick_direction)
		self._selected = None
		self._eh_ids = []


	def create_parent(self):
		self._text = Gtk.Label()
		self._buttons = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
		dialog = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		dialog.pack_start(self._text, True, True, 0)
		dialog.pack_start(self._buttons, True, True, 0)

		dialog.set_name("osd-dialog")
		self._buttons.set_name("osd-dialog-buttons")
		self._text.set_name("osd-dialog-text")
		return dialog


	def pack_items(self, parent, items):
		for item in items:
			if hasattr(item.widget, "set_alignment"):
				item.widget.set_alignment(0.5, 0.5)
			self._buttons.pack_end(item.widget, True, True, 0)


	def use_daemon(self, d):
		"""
		Allows (re)using already existing DaemonManager instance in same process.
		use_config() should be be called before parse_argumets() if this is used.
		"""
		self.daemon = d
		self._connect_handlers()
		self.on_daemon_connected(self.daemon)


	def use_config(self, c):
		"""
		Allows reusing already existin Config instance in same process.
		Has to be called before parse_argumets()
		"""
		self.config = c


	def get_menuid(self):
		# Just to be compatibile with menus when called from scc-osd-daemon
		return None


	def get_selected_item_id(self):
		"""
		Returns ID of selected item or None if nothing is selected.
		"""
		if self._selected:
			return self._selected.id
		return None


	def _add_arguments(self):
		OSDWindow._add_arguments(self)
		self.argparser.add_argument('--confirm-with', type=str,
			metavar="button", default=DEFAULT,
			help="button used to confirm choice")
		self.argparser.add_argument('--cancel-with', type=str,
			metavar="button", default=DEFAULT,
			help="button used to cancel dialog")
		self.argparser.add_argument('--feedback-amplitude', type=int,
			help="enables and sets power of feedback effect generated when active menu option is changed")
		self.argparser.add_argument('--text', type=str, metavar='text',
			help="Dialog text")
		self.argparser.add_argument('items', type=str, nargs='*', metavar='id text',
			help="Dialog buttons")


	def parse_argumets(self, argv):
		if not OSDWindow.parse_argumets(self, argv):
			return False
		if not self.config:
			self.config = Config()

		try:
			self.items = MenuData.from_args(self.args.items)
			self._menuid = None
		except ValueError:
			print('%s: error: invalid number of arguments' % (sys.argv[0]), file=sys.stderr)
			return False

		self._text.set_label(self.args.text)

		if self.args.feedback_amplitude:
			side = "LEFT"
			self.feedback = side, int(self.args.feedback_amplitude)

		# Create buttons that are displayed on screen
		items = self.items.generate(self)
		self.items = []
		for item in items:
			item.widget = self.generate_widget(item)
			if item.widget is not None:
				self.items.append(item)
		self.pack_items(self.parent, self.items)
		if len(self.items) == 0:
			print('%s: error: no items in menu' % (sys.argv[0]), file=sys.stderr)
			return False

		return True


	def generate_widget(self, item):
		""" Generates gtk widget for specified menutitem """
		widget = Gtk.Button.new_with_label(item.label)
		widget.set_relief(Gtk.ReliefStyle.NONE)
		if hasattr(widget.get_children()[0], "set_xalign"):
			widget.get_children()[0].set_xalign(0)
		else:
			widget.get_children()[0].set_halign(Gtk.Align.START)
		widget.set_name("osd-menu-item")

		return widget


	def select(self, index):
		if self._selected:
			self._selected.widget.set_name(self._selected.widget.get_name()
				.replace("-selected", ""))
		if self.items[index].id:
			if self._selected != self.items[index]:
				if self.feedback and self.controller:
					self.controller.feedback(*self.feedback)
			self._selected = self.items[index]
			self._selected.widget.set_name(
					self._selected.widget.get_name() + "-selected")
			return True
		return False


	def _connect_handlers(self):
		self._eh_ids += [
			(self.daemon, self.daemon.connect('dead', self.on_daemon_died)),
			(self.daemon, self.daemon.connect('error', self.on_daemon_died)),
			(self.daemon, self.daemon.connect('alive', self.on_daemon_connected)),
		]


	def run(self):
		self.daemon = DaemonManager()
		self._connect_handlers()
		OSDWindow.run(self)


	def show(self, *a):
		if not self.select(0):
			self.next_item(1)
		OSDWindow.show(self, *a)


	def on_daemon_died(self, *a):
		log.error("Daemon died")
		self.quit(2)


	def on_failed_to_lock(self, error):
		log.error("Failed to lock input: %s", error)
		self.quit(3)


	def on_daemon_connected(self, *a):
		def success(*a):
			log.error("Sucessfully locked input")

		if not self.config:
			self.config = Config()
		self.controller = self.choose_controller(self.daemon)
		if self.controller is None or not self.controller.is_connected():
			# There is no controller connected to daemon
			self.on_failed_to_lock("Controller not connected")
			return

		ccfg = self.config.get_controller_config(self.controller.get_id())
		self._control_with = ccfg["menu_control"]
		self._confirm_with = ccfg["menu_confirm"] if self.args.confirm_with == DEFAULT else self.args.confirm_with
		self._cancel_with = ccfg["menu_cancel"] if self.args.cancel_with == DEFAULT else self.args.cancel_with

		self._eh_ids += [ (self.controller, self.controller.connect('event', self.on_event)) ]
		locks = [ self._control_with, self._confirm_with, self._cancel_with ]
		self.controller.lock(success, self.on_failed_to_lock, *locks)


	def quit(self, code=-2):
		if self.get_controller():
			self.get_controller().unlock_all()
		for source, eid in self._eh_ids:
			source.disconnect(eid)
		self._eh_ids = []
		OSDWindow.quit(self, code)


	def next_item(self, direction):
		""" Selects next menu item, based on self._direction """
		start, i = -1, 0
		try:
			start = self.items.index(self._selected)
			i = start + direction
		except: pass
		while True:
			if i == start:
				# Cannot find valid menu item
				self.select(start)
				break
			if i >= len(self.items):
				i = 0
				continue
			if i < 0:
				i = len(self.items) - 1
				continue
			if self.select(i): break
			i += direction
			if start < 0: start = 0


	def on_stick_direction(self, trash, x, y):
		if x != 0:
			self.next_item(x)


	def on_event(self, daemon, what, data):
		if what == self._control_with:
			self._scon.set_stick(*data)
		elif what == self._cancel_with:
			if data[0] == 0:	# Button released
				self.quit(-1)
		elif what == self._confirm_with:
			if data[0] == 0:	# Button released
				if self._selected and self._selected.callback:
					self._selected.callback(self, self.daemon, self.controller, self._selected)
				elif self._selected:
					self.quit(0)
				else:
					self.quit(-1)



#!/usr/bin/env python3
"""
SC-Controller - Controller Registration Constants

Just huge chunk of constants put aside to make impotant code more readable
"""
from scc.constants import SCButtons, STICK, LEFT, RIGHT
from scc.gui import BUTTON_ORDER

X = 0
Y = 1

AXIS_ORDER = (
	("stick_x", X), ("stick_y", Y),
	("rpad_x", X),  ("rpad_y", Y),
	("lpad_x", X),  ("lpad_y", Y),
	("ltrig", X),	# index 6
	("rtrig", X),
)

STICK_PAD_AREAS = {
	# Numbers here are indexes to AXIS_ORDER tuple
	"STICK":	(STICK, (0, 1)),
	"RPAD":		(RIGHT, (2, 3)),
	"LPAD":		(LEFT, (4, 5)),
}

TRIGGER_AREAS = {
	# Numbers here are indexes to AXIS_ORDER tuple
	"LT": 6,
	"RT": 7
}

AXIS_TO_BUTTON = {
	# Maps stick and dpad axes to their respective "pressed" button
	"stick_x":	SCButtons.STICKPRESS,
	"stick_y":	SCButtons.STICKPRESS,
	"rpad_x":	SCButtons.RPAD,
	"rpad_y":	SCButtons.RPAD,
	"lpad_x":	SCButtons.LPAD,
	"lpad_y":	SCButtons.LPAD,
}

SDL_TO_SCC_NAMES = {
	'guide':			'C',
	'leftstick':		'STICKPRESS',
	'rightstick':		'RPAD',
	'leftshoulder':		'LB',
	'rightshoulder':	'RB',
}

SDL_AXES = (
	# This tuple has to use same order as AXIS_ORDER
	'leftx', 'lefty',
	'rightx', 'righty',
	"dpadx", "dpady",
	'lefttrigger',
	'righttrigger'
)


SDL_DPAD = {
	# Numbers here are indexes to AXIS_ORDER tuple
	# Booleans here are True for positive movements (down/right) and
	# False for negative (up/left)
	'dpdown':	(5, True),
	'dpleft':	(4, False),
	'dpright':	(4, True),
	'dpup':		(5, False),
}



#!/usr/bin/env python3
"""
SC Controller - Special Actions

Special Action is "special" since it cannot be handled by mapper alone.
Instead, on_sa_<actionname> method on handler instance set by
mapper.set_special_actions_handler() is called to do whatever action is supposed
to do. If handler is not set, or doesn't have reqiuired method defined,
action only prints warning to console.
"""
from scc.constants import FE_STICK, FE_TRIGGER, FE_PAD, SCButtons
from scc.constants import LEFT, RIGHT, STICK, SAME
from scc.constants import STICK_PAD_MAX, DEFAULT
from scc.actions import Action, NoAction, SpecialAction, ButtonAction
from scc.actions import HapticEnabledAction, OSDEnabledAction
from scc.actions import MOUSE_BUTTONS
from scc.tools import strip_gesture, nameof, clamp
from scc.modifiers import Modifier, NameModifier
from difflib import get_close_matches
from math import sqrt

import sys, time, logging
log = logging.getLogger("SActions")
_ = lambda x : x


class ChangeProfileAction(Action, SpecialAction):
	SA = COMMAND = "profile"

	def __init__(self, profile):
		Action.__init__(self, profile)
		self.profile = profile


	def describe(self, context):
		if self.name: return self.name
		if context == Action.AC_OSD:
			return _("Profile: %s") % (self.profile,)
		if context == Action.AC_SWITCHER:
			return _("Switch to %s") % (self.profile,)
		return _("Profile Change")


	def get_compatible_modifiers(self):
		return Action.MOD_OSD


	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "%s('%s')" % (self.COMMAND, self.profile)


	def button_release(self, mapper):
		# Execute only when button is released (executing this when button
		# is pressed would send following button_release event to another
		# action from loaded profile)
		self.execute(mapper)


	def whole(self, mapper, *a):
		self.execute(mapper)


class ShellCommandAction(Action, SpecialAction):
	SA = COMMAND = "shell"

	def __init__(self, command):
		#if type(command) == str:
		#	command = command.decode("unicode_escape")
		#assert type(command) == unicode
		Action.__init__(self, command)
		self.command = command


	def describe(self, context):
		if self.name: return self.name
		return _("Execute Command")


	def get_compatible_modifiers(self):
		return Action.MOD_OSD


	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "%s('%s')" % (self.COMMAND, self.parameters[0])


	def button_press(self, mapper):
		# Executes only when button is pressed
		return self.execute(mapper)


class TurnOffAction(Action, SpecialAction):
	SA = COMMAND = "turnoff"

	def __init__(self):
		Action.__init__(self)

	def describe(self, context):
		if self.name: return self.name
		if context == Action.AC_OSD:
			return _("Turning controller OFF")
		return _("Turn Off the Controller")


	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "%s()" % (self.COMMAND,)


	def get_compatible_modifiers(self):
		return Action.MOD_OSD


	def button_release(self, mapper):
		# Execute only when button is released (executing this when button
		# is pressed would hold stuck any other action bound to same button,
		# as button_release is not sent after controller turns off)
		self.execute(mapper)


	def whole(self, mapper, *a):
		self.execute(mapper)


class RestartDaemonAction(Action, SpecialAction):
	SA = COMMAND = "restart"
	ALIASES = ("exit", )

	def __init__(self):
		Action.__init__(self)


	def describe(self, context):
		if self.name: return self.name
		return _("Restart SCC-Daemon")


	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "%s()" % (self.COMMAND,)


	def button_release(self, mapper):
		# Execute only when button is released (for same reason as
		# TurnOffAction does)
		self.execute(mapper)


class LedAction(Action, SpecialAction):
	SA = COMMAND = "led"

	def __init__(self, brightness):
		Action.__init__(self, brightness)
		self.brightness = clamp(0, int(brightness), 100)


	def describe(self, context):
		if self.name: return self.name
		return _("Set LED brightness")


	def get_compatible_modifiers(self):
		return Action.MOD_OSD


	def button_press(self, mapper):
		# Execute only when button is pressed
		self.execute(mapper)


class OSDAction(Action, SpecialAction):
	"""
	Displays text in OSD, or, if used as modifier, displays action description
	and executes that action.
	"""
	SA = COMMAND = "osd"
	DEFAULT_TIMEOUT = 5
	DEFAULT_SIZE = 3
	PROFILE_KEY_PRIORITY = -5	# After XYAction, but beforee everything else

	def __init__(self, *parameters):
		Action.__init__(self, *parameters)
		self.action = None
		self.timeout = self.DEFAULT_TIMEOUT
		self.size = self.DEFAULT_SIZE
		if len(parameters) > 1 and type(parameters[0]) in (int, float):
			# timeout parameter included
			self.timeout = float(parameters[0])
			parameters = parameters[1:]
		if len(parameters) > 1 and type(parameters[0]) in (int, float):
			# size parameter included
			self.size = int(parameters[0])
			parameters = parameters[1:]
		if isinstance(parameters[0], Action):
			self.action = parameters[0]
			self.text = self.action.describe(Action.AC_OSD)
		else:
			self.text = str(parameters[0])
		if self.action and isinstance(self.action, OSDEnabledAction):
			self.action.enable_osd(self.timeout)


	def get_compatible_modifiers(self):
		if self.action:
			return self.action.get_compatible_modifiers()
		return 0


	@staticmethod
	def decode(data, a, *b):
		a = OSDAction(a)
		if data["osd"] is not True:
			a.timeout = float(data["osd"])
		return a


	def describe(self, context):
		if self.name: return self.name
		if self.action:
			return _("%s (with OSD)") % (self.action.describe(context), )
		elif context == Action.AC_OSD:
			return _("Display '%s'" % self.text)
		return _("OSD Message")


	def to_string(self, multiline=False, pad=0):
		parameters = []
		if self.timeout != self.DEFAULT_TIMEOUT or self.size != self.DEFAULT_SIZE:
			parameters.append(str(self.timeout))
		if self.size != self.DEFAULT_SIZE:
			parameters.append(str(self.size))
		if self.action:
			parameters.append(self.action.to_string(multiline=multiline, pad=pad))
		else:
			parameters.append("'%s'" % (str(self.text),))
		return (" " * pad) + "%s(%s)" % (self.COMMAND, ",".join(parameters))


	def strip(self):
		if self.action:
			return self.action.strip()
		return self


	def compress(self):
		if self.action:
			if isinstance(self.action, OSDEnabledAction):
				return self.action.compress()
			self.action = self.action.compress()
		return self


	def button_press(self, mapper):
		self.execute(mapper)
		if self.action:
			return self.action.button_press(mapper)


	def button_release(self, mapper):
		if self.action:
			return self.action.button_release(mapper)


	def trigger(self, mapper, position, old_position):
		if self.action:
			return self.action.trigger(mapper, position, old_position)

	def axis(self, mapper, position, what):
		if self.action:
			return self.action.axis(mapper, position, what)

	def pad(self, mapper, position, what):
		if self.action:
			return self.action.pad(mapper, position, what)

	def whole(self, mapper, x, y, what):
		if self.action:
			return self.action.whole(mapper, x, y, what)


class ClearOSDAction(Action, SpecialAction):
	"""
	Clears all windows from OSD layer. Cancels all menus, clears all messages,
	etc, etc.
	"""
	SA = COMMAND = "clearosd"

	def describe(self, context):
		return _("Hide all OSD Menus and Messages")


	def button_press(self, mapper):
		self.execute(mapper)


class MenuAction(Action, SpecialAction, HapticEnabledAction):
	"""
	Displays menu defined in profile or globally.
	"""
	SA = COMMAND = "menu"
	MENU_TYPE = "menu"
	MIN_STICK_DISTANCE = STICK_PAD_MAX / 3
	DEFAULT_POSITION = 10, -10

	def __init__(self, menu_id, control_with=DEFAULT, confirm_with=DEFAULT,
					cancel_with=DEFAULT, show_with_release=False, size = 0):
		if control_with == SAME:
			# Little touch of backwards compatibility
			control_with, confirm_with = DEFAULT, SAME
		if type(control_with) == int:
			# Allow short form in case when menu is assigned to pad
			# eg.: menu("some-id", 3) sets size to 3
			control_with, size = DEFAULT, control_with
		Action.__init__(self, menu_id, control_with, confirm_with, cancel_with, show_with_release, size)
		HapticEnabledAction.__init__(self)
		self.menu_id = menu_id
		self.control_with = control_with
		self.confirm_with = confirm_with
		self.cancel_with = cancel_with
		self.size = size
		self.x, self.y = MenuAction.DEFAULT_POSITION
		self.show_with_release = bool(show_with_release)
		self._stick_distance = 0


	def describe(self, context):
		if self.name: return self.name
		return _("Menu")


	def get_compatible_modifiers(self):
		return Action.MOD_FEEDBACK


	def to_string(self, multiline=False, pad=0):
		if self.control_with == DEFAULT:
			dflt = (DEFAULT, DEFAULT, False)
			vals = (self.confirm_with, self.cancel_with, self.show_with_release)
			if dflt == vals:
				# Special case when menu is assigned to pad
				if self.size == 0:
					return "%s%s('%s')" % (" " * pad, self.COMMAND, self.menu_id)
				else:
					return "%s%s('%s', %s)" % (" " * pad, self.COMMAND, self.menu_id, self.size)

		return "%s%s(%s)" % (
			" " * pad,
			self.COMMAND,
			",".join(Action.encode_parameters(self.strip_defaults()))
		)


	def get_previewable(self):
		return True


	def button_press(self, mapper):
		if not self.show_with_release:
			confirm_with = self.confirm_with
			cancel_with = self.cancel_with
			args = [ mapper ]
			if confirm_with == SAME:
				confirm_with = mapper.get_pressed_button() or DEFAULT
			elif confirm_with == DEFAULT:
				confirm_with = DEFAULT
			if cancel_with == DEFAULT:
				cancel_with = DEFAULT
			if nameof(self.control_with) in (LEFT, RIGHT):
				args += [ '--use-cursor' ]
			args += [
				'--control-with', nameof(self.control_with),
				'-x', str(self.x), '-y', str(self.y),
				'--size', str(self.size),
				'--confirm-with', nameof(confirm_with),
				'--cancel-with', nameof(cancel_with)
			]
			self.execute(*args)


	def button_release(self, mapper):
		if self.show_with_release:
			self.execute(mapper, '-x', str(self.x), '-y', str(self.y))


	def whole(self, mapper, x, y, what, *params):
		if x == 0 and y == 0:
			# Sent when pad is released - don't display menu then
			return
		if self.haptic:
			params = list(params) + [
				"--feedback-amplitude",
				str(self.haptic.get_amplitude())
			]
		if what in (LEFT, RIGHT):
			confirm_with = self.confirm_with
			cancel_with = self.cancel_with
			if what == LEFT:
				if confirm_with == DEFAULT: confirm_with = SCButtons.LPAD
				if cancel_with == DEFAULT:  cancel_with  = SCButtons.LPADTOUCH
			elif what == RIGHT:
				if confirm_with == DEFAULT: confirm_with = SCButtons.RPAD
				if cancel_with == DEFAULT:  cancel_with  = SCButtons.RPADTOUCH
			else:
				# Stick
				if confirm_with == DEFAULT: confirm_with = SCButtons.STICKPRESS
				if cancel_with == DEFAULT:  cancel_with  = SCButtons.B
			if not mapper.was_pressed(cancel_with):
				self.execute(mapper,
					'--control-with', what,
					'-x', str(self.x), '-y', str(self.y),
					'--use-cursor',
					'--size', str(self.size),
					'--confirm-with', nameof(confirm_with),
					'--cancel-with', nameof(cancel_with),
					*params
				)
		if what == STICK:
			# Special case, menu is displayed only if is moved enought
			distance = sqrt(x*x + y*y)
			if self._stick_distance < MenuAction.MIN_STICK_DISTANCE and distance > MenuAction.MIN_STICK_DISTANCE:
				self.execute(mapper,
					'--control-with', STICK,
					'-x', str(self.x), '-y', str(self.y),
					'--use-cursor',
					'--size', str(self.size),
					'--confirm-with', "STICKPRESS",
					'--cancel-with', STICK,
					*params
				)
			self._stick_distance = distance


class HorizontalMenuAction(MenuAction):
	"""
	Same as menu, but packed as row
	"""
	COMMAND = "hmenu"
	MENU_TYPE = "hmenu"


class GridMenuAction(MenuAction):
	"""
	Same as menu, but displayed in grid
	"""
	COMMAND = "gridmenu"
	MENU_TYPE = "gridmenu"


class QuickMenuAction(MenuAction):
	"""
	Quickmenu. Max.6 items, controller by buttons
	"""
	COMMAND = "quickmenu"
	MENU_TYPE = "quickmenu"


	def describe(self, context):
		if self.name: return self.name
		return _("QuickMenu")


	def button_press(self, mapper):
		# QuickMenu is always shown with release
		pass


	def button_release(self, mapper):
		self.execute(mapper, '-x', str(self.x), '-y', str(self.y))


class RadialMenuAction(MenuAction):
	"""
	Same as grid menu, which is same as menu but displayed in grid,
	but displayed as circle.
	"""
	COMMAND = "radialmenu"
	MENU_TYPE = "radialmenu"

	def __init__(self, menu_id, control_with=DEFAULT, confirm_with=DEFAULT,
					cancel_with=DEFAULT, show_with_release=False, size = 0):
		MenuAction.__init__(self, menu_id, control_with, confirm_with,
						cancel_with, show_with_release, size)
		self.rotation = 0


	def whole(self, mapper, x, y, what):
		if self.rotation:
			MenuAction.whole(self, mapper, x, y, what, "--rotation", self.rotation)
		else:
			MenuAction.whole(self, mapper, x, y, what)


	def set_rotation(self, angle):
		self.rotation = angle


	def get_compatible_modifiers(self):
		return MenuAction.get_compatible_modifiers(self) or Action.MOD_ROTATE


class DialogAction(Action, SpecialAction):
	"""
	Dialog is actually kind of menu, but options for it are different.
	"""
	SA = COMMAND = "dialog"
	DEFAULT_POSITION = 10, -10

	def __init__(self, *pars):
		Action.__init__(self, pars)

		self.options = []
		self.confirm_with = DEFAULT
		self.cancel_with  = DEFAULT
		self.text = _("Dialog")
		self.x, self.y = MenuAction.DEFAULT_POSITION
		# First and 2nd parameter may be confirm and cancel button
		if len(pars) > 0 and pars[0] in SCButtons.__members__.values():
			self.confirm_with, pars = pars[0], pars[1:]
			if len(pars) > 0 and pars[0] in SCButtons.__members__.values():
				self.cancel_with, pars = pars[0], pars[1:]
		# 1st always present argument is title
		if len(pars) > 0:
			self.text, pars = pars[0], pars[1:]
		# ... everything else are actions
		self.options = pars


	def describe(self, context):
		if self.name: return self.name
		return _("Dialog")


	def to_string(self, multiline=False, pad=0):
		rv = "%s%s(" % (" " * pad, self.COMMAND)
		if self.confirm_with != DEFAULT:
			rv += "%s, " % (nameof(self.confirm_with),)
			if self.cancel_with != DEFAULT:
				rv += "%s, " % (nameof(self.cancel_with),)
		rv += "'%s', " % (self.text,)
		if multiline:
			rv += "\n%s" % (" " * (pad + 2))
		for option in self.options:
			rv += "%s, " % (option.to_string(False),)
			if multiline:
				rv += "\n%s" % (" " * (pad + 2))

		rv = rv.strip("\n ,")
		if multiline:
			rv += "\n)"
		else:
			rv += ")"
		return rv


	def get_previewable(self):
		return False


	def button_release(self, mapper):
		confirm_with = self.confirm_with
		cancel_with = self.cancel_with
		args = [
			mapper,
			'-x', str(self.x), '-y', str(self.y),
			'--confirm-with', nameof(confirm_with),
			'--cancel-with', nameof(cancel_with),
			'--text', self.text,
		]
		for x in self.options:
			args.append(x)
		self.execute(*args)


class KeyboardAction(Action, SpecialAction):
	"""
	Shows OSD keyboard.
	"""
	SA = COMMAND = "keyboard"

	def __init__(self):
		Action.__init__(self)


	def get_compatible_modifiers(self):
		return Action.MOD_POSITION


	def describe(self, context):
		if self.name: return self.name
		if context == Action.AC_OSD:
			return _("Display Keyboard")
		return _("OSD Keyboard")


	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "%s()" % (self.COMMAND,)


	def button_press(self, mapper):
		self.execute(mapper)


class PositionModifier(Modifier):
	"""
	Sets position for OSD menu.
	"""
	COMMAND = "position"

	def _mod_init(self, x, y):
		self.position = (x, y)


	def compress(self):
		if isinstance(self.action, MenuAction):
			self.action.x, self.action.y = self.position
		return self.action


	@staticmethod
	def decode(data, a, *b):
		x, y = data[PositionModifier.COMMAND]
		return PositionModifier(x, y, a)


	def describe(self, context):
		return self.action.describe(context)


class GesturesAction(Action, OSDEnabledAction, SpecialAction):
	"""
	Stars gesture detection on pad. Recognition is handled by whatever
	is special_actions_handler and results are then sent back to this action
	as parameter of gesture() method.
	"""
	SA = COMMAND = "gestures"
	PROFILE_KEYS = ("gestures",)
	PROFILE_KEY_PRIORITY = 2
	DEFAULT_PRECISION = 1.0

	def __init__(self, *stuff):
		OSDEnabledAction.__init__(self)
		Action.__init__(self, *stuff)
		self.gestures = {}
		self.precision = self.DEFAULT_PRECISION
		gstr = None

		if len(stuff) > 0 and type(stuff[0]) in (int, float):
			self.precision = clamp(0.0, float(stuff[0]), 1.0)
			stuff = stuff[1:]

		for i in stuff:
			if gstr is None and type(i) == str:
				gstr = i
			elif gstr is not None and isinstance(i, Action):
				self.gestures[gstr] = i
				gstr = None
			else:
				raise ValueError("Invalid parameter for '%s': unexpected %s" % (
						self.COMMAND, i))


	def get_compatible_modifiers(self):
		return Action.MOD_OSD


	def describe(self, context):
		if self.name: return self.name
		return _("Gestures")


	def to_string(self, multiline=False, pad=0):
		if multiline:
			rv = [ (" " * pad) + self.COMMAND + "(" ]
			if self.precision != self.DEFAULT_PRECISION:
				rv[0] += "%s," % (self.precision)
			for gstr in self.gestures:
				a_str = self.gestures[gstr].to_string(True).split("\n")
				a_str[0] = (" " * pad) + "  '" + (gstr + "',").ljust(11) + a_str[0]	# Key has to be one of SCButtons
				for i in range(1, len(a_str)):
					a_str[i] = (" " * pad) + "  " + a_str[i]
				a_str[-1] = a_str[-1] + ","
				rv += a_str
			if rv[-1][-1] == ",":
				rv[-1] = rv[-1][0:-1]
			rv += [ (" " * pad) + ")" ]
			return "\n".join(rv)
		else:
			rv = [ ]
			if self.precision != self.DEFAULT_PRECISION:
				rv.append(str(self.precision))
			for gstr in self.gestures:
				rv += [ "'%s'" % (gstr,), self.gestures[gstr].to_string(False) ]
			return self.COMMAND + "(" + ", ".join(rv) + ")"


	def compress(self):
		for gstr in self.gestures:
			a = self.gestures[gstr].compress()
			if "i" in gstr:
				del self.gestures[gstr]
				gstr = strip_gesture(gstr)
			self.gestures[gstr] = a
		return self


	@staticmethod
	def decode(data, a, parser, *b):
		args = []
		ga = GesturesAction()
		ga.gestures = {
			gstr : parser.from_json_data(data[GesturesAction.PROFILE_KEYS[0]][gstr])
			for gstr in data[GesturesAction.PROFILE_KEYS[0]]
		}
		if "name" in data:
			ga.name = data["name"]
		if "osd" in data:
			ga = OSDAction(ga)
		return ga

	def _find_exact_gesture(self, gesture_string):
		return self.gestures.get(gesture_string)

	def _find_ignore_stroke_count_gesture(self, gesture_string):
		stripped_gesture_string = strip_gesture(gesture_string)
		return self.gestures.get(stripped_gesture_string)

	def _find_best_match_gesture(self, gesture_string):
		NUM_MATCHES_TO_RETURN = 1

		similar_gestures = get_close_matches(gesture_string, self.gestures.keys(), NUM_MATCHES_TO_RETURN, self.precision)
		best_gesture = next(iter(similar_gestures), None)

		if best_gesture is not None:
			return self.gestures[best_gesture]
		else:
			return None

	def find_gesture_action(self, gesture_string):
		action = None
		action = action or self._find_exact_gesture(gesture_string)
		action = action or self._find_ignore_stroke_count_gesture(gesture_string)
		action = action or self._find_best_match_gesture(gesture_string)
		return action

	def gesture(self, mapper, gesture_string):
		action = self.find_gesture_action(gesture_string)
		if action:
			action.button_press(mapper)
			mapper.schedule(0, action.button_release)

	def whole(self, mapper, x, y, what):
		if (x, y) != (0, 0):
			# (0, 0) singlanizes released touchpad
			self.execute(mapper, x, y, what)


class CemuHookAction(Action, SpecialAction):
	SA = COMMAND = "cemuhook"
	MAGIC_GYRO = (2000.0 / 32768.0)
	ACC_RES_PER_G = 16384.0

	def gyro(self, mapper, *pyr):
		sa_data = (
			-mapper.state.accel_x / CemuHookAction.ACC_RES_PER_G, # AccelX
			-mapper.state.accel_z / CemuHookAction.ACC_RES_PER_G, # AccelZ
			mapper.state.accel_y / CemuHookAction.ACC_RES_PER_G, # AccelY
			pyr[0] * CemuHookAction.MAGIC_GYRO, # Gyro Pitch
			-pyr[1] * CemuHookAction.MAGIC_GYRO, # Gyro Yaw
			-pyr[2] * CemuHookAction.MAGIC_GYRO, # Gyro Roll
		)
		#log.debug(sa_data)
		self.execute(mapper, sa_data)

	def describe(self, context):
		if self.name: return self.name
		return _("CemuHook")

# Register actions from current module
Action.register_all(sys.modules[__name__])



#!/usr/bin/env python3
import signal
import sys

def main() -> None:
	def sigint(*a):
		print("\n*break*")
		sys.exit(0)

	if __name__ == "__main__":
		signal.signal(signal.SIGINT, sigint)

		import gi
		gi.require_version('Gtk', '3.0')
		gi.require_version('Rsvg', '2.0')
		gi.require_version('GdkX11', '3.0')

		from scc.tools import init_logging
		from scc.paths import get_share_path
		init_logging()

		from scc.osd.message import Message
		m = Message()
		if not m.parse_argumets(sys.argv):
			sys.exit(1)
		m.run()
		sys.exit(m.get_exit_code())

if __name__ == "__main__":
	main()



#!/usr/bin/env python3
"""
SC-Controller - GDK_TO_KEY

Maps Gdk.KEY_* constants into Keys.KEY_* constants.
Used by ActionEditor (when grabbing the key)
"""
from gi.repository import Gdk
from scc.uinput import Keys

GDK_TO_KEY = {
	# Row 1
	Gdk.KEY_Escape		: Keys.KEY_ESC,
	Gdk.KEY_Print		: Keys.KEY_PRINT,
	Gdk.KEY_Scroll_Lock	: Keys.KEY_SCROLLLOCK,
	Gdk.KEY_Sys_Req		: Keys.KEY_SYSRQ,
	Gdk.KEY_Pause		: Keys.KEY_PAUSE,

	# Row 2
	Gdk.KEY_quoteleft	: Keys.KEY_GRAVE,	# tilde
	Gdk.KEY_minus		: Keys.KEY_MINUS,
	Gdk.KEY_equal		: Keys.KEY_EQUAL,
	Gdk.KEY_BackSpace	: Keys.KEY_BACKSPACE,

	# Row 3
	Gdk.KEY_Tab			: Keys.KEY_TAB,
	Gdk.KEY_bracketleft	: Keys.KEY_LEFTBRACE,
	Gdk.KEY_bracketright: Keys.KEY_RIGHTBRACE,
	Gdk.KEY_backslash	: Keys.KEY_BACKSLASH,

	# Row 4
	Gdk.KEY_Caps_Lock	: Keys.KEY_CAPSLOCK,
	Gdk.KEY_semicolon	: Keys.KEY_SEMICOLON,
	Gdk.KEY_apostrophe	: Keys.KEY_APOSTROPHE,
	Gdk.KEY_Return		: Keys.KEY_ENTER,

	# Row 5
	Gdk.KEY_Shift_L		: Keys.KEY_LEFTSHIFT,
	Gdk.KEY_comma		: Keys.KEY_COMMA,
	Gdk.KEY_period		: Keys.KEY_DOT,
	Gdk.KEY_slash		: Keys.KEY_SLASH,
	Gdk.KEY_Shift_R		: Keys.KEY_RIGHTSHIFT,

	# Numpad
	Gdk.KEY_KP_0		: Keys.KEY_KP0,
	Gdk.KEY_KP_1		: Keys.KEY_KP1,
	Gdk.KEY_KP_2		: Keys.KEY_KP2,
	Gdk.KEY_KP_3		: Keys.KEY_KP3,
	Gdk.KEY_KP_4		: Keys.KEY_KP4,
	Gdk.KEY_KP_5		: Keys.KEY_KP5,
	Gdk.KEY_KP_6		: Keys.KEY_KP6,
	Gdk.KEY_KP_7		: Keys.KEY_KP7,
	Gdk.KEY_KP_8		: Keys.KEY_KP8,
	Gdk.KEY_KP_9		: Keys.KEY_KP9,
	Gdk.KEY_KP_Delete	: Keys.KEY_KPDOT,
	Gdk.KEY_KP_Divide	: Keys.KEY_KPSLASH,
	Gdk.KEY_KP_Add		: Keys.KEY_KPPLUS,
	Gdk.KEY_KP_Multiply	: Keys.KEY_KPASTERISK,
	Gdk.KEY_KP_Subtract	: Keys.KEY_KPMINUS,
	Gdk.KEY_KP_Enter	: Keys.KEY_KPENTER,
	Gdk.KEY_Num_Lock	: Keys.KEY_NUMLOCK,

	# Home & co
	Gdk.KEY_Insert		: Keys.KEY_INSERT,
	Gdk.KEY_Home		: Keys.KEY_HOME,
	Gdk.KEY_Page_Up		: Keys.KEY_PAGEUP,
	Gdk.KEY_Delete		: Keys.KEY_DELETE,
	Gdk.KEY_End			: Keys.KEY_END,
	Gdk.KEY_Page_Down	: Keys.KEY_PAGEDOWN,

	# Arrows
	Gdk.KEY_Up			: Keys.KEY_UP,
	Gdk.KEY_Left		: Keys.KEY_LEFT,
	Gdk.KEY_Right		: Keys.KEY_RIGHT,
	Gdk.KEY_Down		: Keys.KEY_DOWN,

	# Bottom row
	Gdk.KEY_Control_L	: Keys.KEY_LEFTCTRL,
	Gdk.KEY_Super_L		: Keys.KEY_LEFTMETA,
	Gdk.KEY_Alt_L		: Keys.KEY_LEFTALT,
	Gdk.KEY_space		: Keys.KEY_SPACE,
	Gdk.KEY_Alt_R		: Keys.KEY_RIGHTALT,
	Gdk.KEY_Super_R		: Keys.KEY_RIGHTMETA,
	Gdk.KEY_Menu		: Keys.KEY_COMPOSE,
	Gdk.KEY_Control_R	: Keys.KEY_RIGHTCTRL,
}


KEYCODE_TO_KEY = {
	# Row 1
	9	: Keys.KEY_ESC,
	67	: Keys.KEY_F1,
	68	: Keys.KEY_F2,
	69	: Keys.KEY_F3,
	70	: Keys.KEY_F4,
	71	: Keys.KEY_F5,
	72	: Keys.KEY_F6,
	73	: Keys.KEY_F7,
	74	: Keys.KEY_F8,
	75	: Keys.KEY_F9,
	76	: Keys.KEY_F10,
	95	: Keys.KEY_F11,
	96	: Keys.KEY_F12,
	107	: Keys.KEY_PRINT,
	78	: Keys.KEY_SCROLLLOCK,
	127		: Keys.KEY_PAUSE,

	# Row 2
	49	: Keys.KEY_GRAVE,	# tilde
	10	: Keys.KEY_1,
	11	: Keys.KEY_2,
	12	: Keys.KEY_3,
	13	: Keys.KEY_4,
	14	: Keys.KEY_5,
	15	: Keys.KEY_6,
	16	: Keys.KEY_7,
	17	: Keys.KEY_8,
	18	: Keys.KEY_9,
	19	: Keys.KEY_0,
	20	: Keys.KEY_MINUS,
	21	: Keys.KEY_EQUAL,
	22	: Keys.KEY_BACKSPACE,

	# Row 3
	23	: Keys.KEY_TAB,
	24	: Keys.KEY_Q,
	25	: Keys.KEY_W,
	26	: Keys.KEY_E,
	27	: Keys.KEY_R,
	28	: Keys.KEY_T,
	29	: Keys.KEY_Y,
	30	: Keys.KEY_U,
	31	: Keys.KEY_I,
	32	: Keys.KEY_O,
	33	: Keys.KEY_P,
	34	: Keys.KEY_LEFTBRACE,
	35	: Keys.KEY_RIGHTBRACE,
	51	: Keys.KEY_BACKSLASH,

	# Row 4
	66	: Keys.KEY_CAPSLOCK,
	38	: Keys.KEY_A,
	39	: Keys.KEY_S,
	40	: Keys.KEY_D,
	41	: Keys.KEY_F,
	42	: Keys.KEY_G,
	43	: Keys.KEY_H,
	44	: Keys.KEY_J,
	45	: Keys.KEY_K,
	46	: Keys.KEY_L,
	47	: Keys.KEY_SEMICOLON,
	48	: Keys.KEY_APOSTROPHE,
	36	: Keys.KEY_ENTER,

	# Row 5
	50	: Keys.KEY_LEFTSHIFT,
	52	: Keys.KEY_Z,
	53	: Keys.KEY_X,
	54	: Keys.KEY_C,
	55	: Keys.KEY_V,
	56	: Keys.KEY_B,
	57	: Keys.KEY_N,
	58	: Keys.KEY_M,
	59	: Keys.KEY_COMMA,
	60	: Keys.KEY_DOT,
	61	: Keys.KEY_SLASH,
	62	: Keys.KEY_RIGHTSHIFT,

	# Numpad
	90	: Keys.KEY_KP0,
	87	: Keys.KEY_KP1,
	88	: Keys.KEY_KP2,
	89	: Keys.KEY_KP3,
	83	: Keys.KEY_KP4,
	84	: Keys.KEY_KP5,
	85	: Keys.KEY_KP6,
	79	: Keys.KEY_KP7,
	80	: Keys.KEY_KP8,
	81	: Keys.KEY_KP9,
	91	: Keys.KEY_KPDOT,
	106	: Keys.KEY_KPSLASH,
	86	: Keys.KEY_KPPLUS,
	63	: Keys.KEY_KPASTERISK,
	82	: Keys.KEY_KPMINUS,
	104	: Keys.KEY_KPENTER,
	77	: Keys.KEY_NUMLOCK,

	# Home & co
	118	: Keys.KEY_INSERT,
	110	: Keys.KEY_HOME,
	112	: Keys.KEY_PAGEUP,
	119	: Keys.KEY_DELETE,
	115	: Keys.KEY_END,
	117	: Keys.KEY_PAGEDOWN,

	# Arrows
	111	: Keys.KEY_UP,
	113	: Keys.KEY_LEFT,
	114	: Keys.KEY_RIGHT,
	116	: Keys.KEY_DOWN,

	# Bottom row
	37	: Keys.KEY_LEFTCTRL,
	133	: Keys.KEY_LEFTMETA,
	64	: Keys.KEY_LEFTALT,
	65	: Keys.KEY_SPACE,
	108	: Keys.KEY_RIGHTALT,
	134	: Keys.KEY_RIGHTMETA,
	135	: Keys.KEY_COMPOSE,
	105	: Keys.KEY_RIGHTCTRL,
}

# Stuff that is missing above is auto-generated here
names = { x.name : x for x in Keys }

for x in dir(Gdk):
	if x.startswith("KEY_"):
		if x in names:
			GDK_TO_KEY[getattr(Gdk, x)] = names[x]

# A-Z keys, because GDK has different codes for 'A' and 'a'
for x in range(ord('a'), ord('z')+1):
	GDK_TO_KEY[getattr(Gdk, "KEY_" + chr(x))] = names["KEY_" + chr(x).upper()]

KEY_TO_GDK = { GDK_TO_KEY[a] : a for a in GDK_TO_KEY }
KEY_TO_KEYCODE = { KEYCODE_TO_KEY[a] : a for a in KEYCODE_TO_KEY }

def keyevent_to_key(event):
	if event.hardware_keycode in KEYCODE_TO_KEY:
		return KEYCODE_TO_KEY[event.hardware_keycode]
	return None
